#! /usr/bin/env node
import parseArgs from "minimist"
import path from "path"
import fs from "fs"
import { fileURLToPath } from "url"

const isUserPipingInput = () => {
  if (process.platform === "win32") return false
  if (process.stdin.isTTY === true) return false
  return fs.fstatSync(0).isFIFO()
}

class SimpleCLI {
  CommandFnDecoratorSuffix = "Command"
  verbose = true

  async executeUsersInstructionsFromShell(
    args = parseArgs(process.argv.slice(2))._,
    userIsPipingInput = isUserPipingInput(),
  ) {
    const cwd = process.cwd()
    const command = args[0]
    const commandName = `${command}${this.CommandFnDecoratorSuffix}`
    if (this[commandName])
      return userIsPipingInput
        ? this._runCommandOnPipedStdIn(commandName)
        : this[commandName](cwd, args.slice(1))
    else if (command) this.log(`No command '${command}'. Running help command.`)
    else this.log(`No command provided. Running help command.`)
    return this.helpCommand()
  }

  _runCommandOnPipedStdIn(commandName) {
    this.log(`Running ${commandName} on piped input`)
    let pipedData = ""

    process.stdin.on("readable", function () {
      pipedData += this.read()
    })

    process.stdin.on("end", async () => {
      const folders = pipedData
        .trim()
        .split("\n")
        .map((line) => line.trim())
        .filter((line) => fs.existsSync(line))

      for (const line of folders) {
        await this[commandName](line)
      }

      if (folders.length === 0) {
        this[commandName](process.cwd())
      }
    })
  }

  silence() {
    this.verbose = false
    return this
  }

  log(message) {
    if (this.verbose) console.log(message)
    return message
  }

  get _allCommands() {
    return Object.getOwnPropertyNames(Object.getPrototypeOf(this))
      .filter((atom) => atom.endsWith(this.CommandFnDecoratorSuffix))
      .sort()
  }

  resolvePath(folder = "") {
    return path.isAbsolute(folder)
      ? path.normalize(folder)
      : path.resolve(path.join(process.cwd(), folder))
  }

  helpCommand() {
    this.log(this.welcomeMessage)
    return this.log(
      `\nAvailable commands:\n\n${this._allCommands.map((comm) => `🖌️ ` + comm.replace(this.CommandFnDecoratorSuffix, "")).join("\n")}\n`,
    )
  }
}

export { SimpleCLI }
